{exec} = require 'child_process'
assert = require 'assert'
async = require 'async'
coffee = require 'coffee-script'
errors = []
fs	   = require 'fs'
growl = require 'growl'
http = require 'http'
Log = require 'coloured-log'
path = require 'path'
request = require 'request'
stalker = require 'stalker'
sys = require 'sys'
util   = require 'util'


# Initialize variables used in program execution and logging
config = try require "#{rootDir}/kitkat.config.js" # Overrides configuration functionality
config = config?={noConfig: true, noCompile: true, safetyOff: false}
log = new Log()

# Path Variables
libDir			= path.join(path.dirname(fs.realpathSync(__filename)), '../lib')
rootDir			= process.cwd()

# Global Variables required by kitkat
autoSpec = false
expresso = path.join(path.dirname(fs.realpathSync(__filename)), '../bin/expresso spec/*.coffee -I lib')
file_matcher = /\.(js|coffee)$/
intialTestComplete = false
prodDeployDir	= path.join("#{rootDir}", config.prodDeployDir?="lib")
prodSrcDir		= path.join("#{rootDir}", config.prodSrcDir?="src")		# assumed that code files could be within the project anywhere
specFileDir		= path.join("#{rootDir}", config.specFileDir?="spec")

prodCoffeeFiles = []
specCoffeeFiles = []
failures = 0
testcount = 0
requestCount=0  # Tracks open test cases to assert.response
files = []

prodCoffeeOpts = "--output #{prodDeployDir}"
specCoffeeOpts = "--output #{specFileDir}"

# Set the location to find libraries used by spec files
if config.noCompile==true
	require.paths.unshift("#{prodSrcDir}")
else
	require.paths.unshift("#{prodDeployDir}")

# Take control of stdout for tests to block any messages that do not contain '!!!'
# Unfortunately if a program writes to stderr (like express with console.error) we can not intercept it, those messages will display.
hook_stdout = (callback) ->
	old_write = process.stdout.write
	process.stdout.write = ((write) ->
		(string, encoding, fd) ->
			if string.match(/!!!/)
				string = string.replace("!!!", "")
				write.apply process.stdout, arguments
			callback string, encoding, fd
	)(process.stdout.write)
	->
		process.stdout.write = old_write
		
unhook = null # used to store and reset stdout handling during tests.


process.on "uncaughtException", (err) ->
	log.emergency "Uncaught Exception:  #{err}"

#########################################
# Startup Functions required by kitkat
#########################################

# Verify paths exist, create if they do not, send back error if any issues occured.
checkPaths = (callback) ->
	if config.safetyOff
		callback null # this will cause us not to check or autogenerate content, depending on whether your test cases require it or not.
	else
		path.exists prodDeployDir, (exists) ->
		
			if not exists and config.noCompile==false # If we do not compile we do not care about the deploy location for our code.
				log.warning "The lib path was not found, this has been created."
				exec "mkdir #{prodDeployDir}", (err, stdout, stderr) ->
					if err
						callback err
			path.exists specFileDir, (exists) ->
				if not exists
					log.warning "The spec path was not found, this has been created."
					exec "mkdir #{specFileDir}", (err, stdout, stderr) ->
						if err
							callback err
				path.exists prodSrcDir, (exists) ->
					if not exists
						exec "mkdir #{prodSrcDir}", (err, stdout, stderr) ->
							if err
								callback err
							else
								log.warning "The src path was not found, this has been created and a default app.coffee file created."
								exec "touch #{prodSrcDir}/app.coffee", (err, stdout, stderr) ->
									if not config.noCompile==true
										log.notice "Compiling lib/app.js from default app.coffee"
										exec "coffee --output #{prodDeployDir} --compile #{prodSrcDir}/app.coffee", (err,stdout, stderr) ->
											callback err
									else
										callback err
					else
						callback null
	
getVersion = (callback) ->
	# Get version of installed kitkat application
	fs.readFile path.join(path.dirname(fs.realpathSync(__filename)),"../package.json"), (err, data) ->
		throw err if err
		package = JSON.parse(data)
		callback package

showVersion = (argv) ->
	# Show version of installed kitkat application
	getVersion (package) ->
		log.info colorize("[purple]{Kitkat - #{package.version}}")

startHeader = (action) ->
	log.notice colorize("[purple]{Kitkat - Kontinuous Integrated Testing Koffee Application Template}")
	log.notice colorize("[red]{press CTRL-C to quit}\n\n")
	if config.noCompile==false
		log.notice colorize("[bold]{noCompile:} false - This means you are creating .js files in #{prodDeployDir}")
	if config.safetyOff==true
		log.notice colorize("[bold]{safetyOff:} True")
	log.notice "#{action}"
	log.notice colorize("[bold]{source files:} #{prodSrcDir}")
	log.notice colorize("[bold]{spec files:} #{specFileDir}\n")

# Entry point to kitkat
exports.run = ->
	# Commandline processor
	optimist = require 'optimist'
	args = optimist.argv

	optimist.alias 'test', 't'
	optimist.alias 'spec', 's'
	optimist.alias 'autospec', 'a'
	optimist.alias 'version', 'v'
	optimist.usage(colorize("[purple]{Kitkat - Kontinuous Integrated Testing Koffee Application Template}\n\n [bold]{Default usage} - with no arguments, it will monitor the current project for changes and\n\t\t run all spec files found against all coffee files found."), {
		'test': {description: colorize("[green]{runs all tests against your application files in the spec path\n}"), required: false},
		'spec': {description: colorize("[green]{generates spec files based on the supplied path/file, if none\n\t\t  provided then processes all coffee files in project\n}"), required: false},
		'autospec': {description: colorize("[green]{generates new spec files and monitors all files for changes,\n\t\t  automatically runs tests on startup and as files change\n}"), required: false},
		'version': {description: colorize("[green]{displays version number for kitkat}"), required: false}
		})
	commands = []
	for key, value of args
		try 
			switch key.toLowerCase()
				when "t", "test", "s", "spec", "a", "autotest", "v", "version"
					commands.push key
				else
					if key !="$0" and key !="_"
						throw "Invaid argument"
		catch ex
			log.error "#{ex}\n"
			optimist.showHelp() 
			return
	
	command = commands[0]?.toLowerCase()
	
	if command!='v' and command!= 'version'
		checkPaths (err) ->
			if err
				handleError(err)
				callback err
		
			switch command
				when "t", "test"
					startHeader "Initiating tests"
					tester() # args.t ?= args.test
				when "s", "spec"
					startHeader "Generating spec files"
					spec args.s ?= args.spec
				when "a", "autospec"
					startHeader "Autospec and testing current project"
					autospec()
				else
					startHeader "Monitoring and testing current project"
					main()
	else
		showVersion()

#########################################
# Functions follow in alphabetical order
#########################################

# Override the assert lib AssertionError so we can capture and control how this information is displayed.
assert.AssertionError = AssertionError = (options) ->
	@message = options.message
	@actual = ""
	@expected = ""
	if options.actual?
		@actual = "\n[bold]{Actual:} [red]{#{options.actual}}"
	if options.expected?
		@expected = "\n[bold]{Expected:} [green]{#{options.expected}}"
	log.error colorize("!!![red]{Test failed}\n\nMessage: [bold]{#{@message}}\n#{@actual}#{@expected}\nFailures: #{failures}\n\n[bold]{================================================================}\n")

util.inherits assert.AssertionError, Error

assert.response = (server, req, res, msg) ->
	requestCount++
	requestFailures=0
	url = ""
	if req.secure==true
		url += "https://"
	else
		url += "http://"
	url += "#{server.address().address}:#{server.address().port}#{req.url}"
	request {uri: url}, (error, response, body) ->
		requestCount--
		if res.body != undefined 
			eql = (if res.body instanceof RegExp then res.body.test(response.body) else res.body == response.body)
			try
				assert.ok eql, colorize("#{msg()}[red]{Invalid response body.}\n\n[bold]{Actual:} Please review [red]{#{url}}\n[bold]{Expected:} [green]{#{sys.inspect(res.body)}}\n")
			catch err
				if ++requestFailures <= 1
					failures++
		if typeof response.statusCode == "number"
			try
				if typeof res['status'] == "number"
					assert.equal response.statusCode, res['status'], colorize("#{msg()}[red]{Invalid response status code.}")
			catch err
				if ++requestFailures <= 1
					failures++
		if res.headers
			keys = Object.keys(res.headers)
			i = 0
			len = keys.length
			while i < len
				name = keys[i]
				actual = response.headers[name.toLowerCase()]
				expected = res.headers[name]
				eql = (if expected instanceof RegExp then expected.test(actual) else expected == actual)
				try
					assert.equal actual, expected, colorize("#{msg()}[red]{Invalid response header}")
				catch err
					if ++requestFailures <= 1
						failures++
				++i
	
autospec = ->
	autoSpec = true
	main()

coffee = (options = "", file) ->
	log.info "Compiling #{file}"
	exec "coffee #{options} --compile #{file}", (err, stdout, stderr) -> 
		handleError(err) if err

colorize = (str) ->
	colors = 
		bold: 1
		red: 31
		green: 32
		yellow: 33
		purple: 35
		cyan: 36
	str.replace /\[(\w+)\]\{([^]*?)\}/g, (_, color, str) ->
		"\u001b[" + colors[color] + "m" + str + "\u001b[0m"

displayNotification = (message = '') -> 
	options = {
		title: 'CoffeeScript'
		image: 'lib/CoffeeScript.png'
	}
	growl.notify message, options

gatherFiles = (filepath, extRegex, callback) ->
	filesFound = []
	file = path.normalize(filepath)
	fs.stat file, (err,stats) ->
		if err
			file += extRegex.toString().replace(/\//g,"").replace(/\\/g,"")
			fs.stat file, (err,stats) ->
				callback err, null if err 
				if not err
					filesFound.push file 
					callback null, filesFound
		else
			if stats.isDirectory()
				walk path.normalize(file + '/'), extRegex, (err,results) ->
					callback err, null if err
					if not err
						filesFound = results	
						callback null, filesFound
			else
				filesFound.push file
				callback null, filesFound

handleError = (error) -> 
	log.error error
	displayNotification error

main = ->
	stalker.watch "#{prodSrcDir}" , watchCoffeeFile, unWatchCoffeeFile
	stalker.watch "#{specFileDir}" , watchCoffeeFile, unWatchCoffeeFile
	tester()

report = (title) ->
	log.notice "#{failures}"
	status = 'green'
	message = ""
	options = {}
	sys.print colorize("[bold]{\n=======================================================================================\nTesting of '#{title}' completed.\n\n}")
	if failures
		status = 'red'
		options = {
			title: 'Automated Testing'
			image: "#{libDir}/error.png"
			}
		message = "Test Failure\n\n#{failures} out of #{testcount} failed.\n\nPlease review console for details." 
		log.error colorize("[bold]{Test Results:} [#{status}]{#{testcount-failures} of #{testcount} passed.}")
	else
		options = {
			title: 'Automated Testing'
			image: "#{libDir}/success.png"
			}
		message = "All Tests Passed\n\n#{testcount} test(s) completed."
		log.info colorize("[bold]{Test Results:} [#{status}]{#{testcount-failures} of #{testcount} passed.}")
		
	growl.notify message, options
	failures = 0	#Reset counters at end of testing for next test
	testcount = 0

runTests = (callback) ->
	files = []
	try
		files = fs.readdirSync("#{specFileDir}")
	catch err
		log.emergency "Failed to load tests in [bold]{#{specFileDir}}\n"
		++failures
		process.exit 1
	files.forEach runTestFile
	callback null

runTestFile = (file) ->
	require('sys').print colorize("[purple]{\n=======================================================================================\n}")
	log.notice colorize("[purple]{#{file}} being processed....")
	if file.match(file_matcher)
		title = path.basename(file)
		file = path.join("#{specFileDir}", file)
		mod = require(file)
		(check = ->
			len = Object.keys(mod).length
			if len
				runTestSuite title, mod
			else
				setTimeout check, 20
		)()
		delete require.cache[file] # removes the test mod from cache so it is reloaded next time
		
runTestSuite = (title, tests) ->
	console.log "#{failures} initial"
	unhook = hook_stdout((string, encoding, fd) ->
	)
	keys = Object.keys(tests)
	setup = {}
	teardown = {}
	try
		setup = tests['setup']
		if typeof setup is "function" 
			log.notice "!!!Suite contains setup, function completed successfully."
			setup()
	catch err
		log.error "!!!Suite setup failed\n\n #{err}\n"
	
	teardown = tests['teardown']
		
	process.setMaxListeners 10 + process.listeners("beforeExit").length + keys.length
	
	finish = ->
		unless requestCount
			try
				if typeof teardown is "function"
					teardown()
					log.notice "!!!Teardown function completed successfully."
			catch err
				log.error "!!!Teardown function failed\n\n#{err}\n"
			unhook()
			report(title)
		else
			process.nextTick finish  # keep checking until requests are done...so we dont shut off server before it is complete.

	(next = ->
		if keys.length
			test = tests[key = keys.shift()]
			return next()	 if key == "setup" or key == "teardown"
			testcount++
			try
				test ->
					console.log "!!!Testdone #{failures}"
					process.on "beforeExit", ->
						# Assert test is completed
				next()
			catch err
				failures++
				next()
		else
			console.log "!!!finsihed #{failures}"
			finish()

	)()


spec = (filepath) ->
	async.series
		gather_files: (callback) ->
			filepath = "/" if (filepath==true)
			filepath = path.join(prodSrcDir, path.normalize(filepath))
			gatherFiles filepath, /\.coffee/, (err, results) ->
				prodCoffeeFiles = results
				callback err, results
		create_specs: (callback) ->
			pending = prodCoffeeFiles.length
			for file, index in prodCoffeeFiles then do (file, index) ->
				specFile = path.join(specFileDir, path.basename(file, '.coffee') + ".spec.coffee")
				fs.stat specFile, (err,stats) ->
					if err # File does not exist
						prodFile = path.join(prodDeployDir, path.basename(file, '.coffee') )
						writeSpecFile specFile, prodFile, (error, results) ->
							callback err, null if error
				callback null, null unless --pending
	, (err, results) ->
		handleError(err) if err

tester = ->
	async.series
		compile_prod: (callback) ->
			if config.noCompile==true
				callback null,null # Some projects do not require us to recompile anything, rather just watch the files for changes
			else
				gatherFiles prodSrcDir, /\.coffee/, (err, results) ->
					prodCoffeeFiles = results
					pending = prodCoffeeFiles.length
					for file, index in prodCoffeeFiles then do (file, index) ->
						exec "coffee --output #{prodDeployDir} --compile #{file}", (err,stdout, stderr) ->
							callback null, null unless --pending

		create_specs: (callback) ->
			if autoSpec==true
				spec "/"
			callback null, null

		run_specs: (callback) ->
			runTests (err) ->
				intialTestComplete = true
				callback err, null

unWatchCoffeeFile = (err, file) ->
	if "#{file}".match(/\.coffee/)?
		util.warning "#{file} removed."
		fs.unwatchFile "#{file}"
		runTests (err) ->
			return

walk = (dir, extRegex, callback) ->
	results = []
	pending = 0
	fs.readdir dir, (err, list) ->
		callback err, null if err
		pending = list.length
		for file in list then do(file) ->
			file = dir + file
			fs.stat file, (err, stat) ->
				if stat and stat.isDirectory()
					file += "/"
					if "#{file}".match(/node_modules/)?
						callback null,results unless --pending
					else
						walk file, extRegex, (err, res) ->
							results = results.concat(res)
							callback null, results unless --pending
				else
					if file.match(extRegex)
						results.push file
					callback null, results unless --pending

watchCoffeeFile = (err, file) ->
	# Test that we are looking at a .coffee file which is not a spec file
	if not "#{file}".match(/\.spec\.coffee/)? and "#{file}".match(/\.coffee/)?

		log.notice "Watch added for " + colorize("[cyan]{#{file}}")
		fs.unwatchFile "#{file}"
		fs.watchFile "#{file}", (curr, prev) ->
			if +curr.mtime isnt +prev.mtime
				log.notice "Saw change in " + colorize("[cyan]{#{file}}")
				if not config.noCompile==true
					exec "coffee --output #{prodDeployDir} --compile #{file}", (err,stdout, stderr) ->
						runTests (err) ->
							return
				else
					runTests (err) ->
						return
	else if "#{file}".match(/\.spec\.coffee/)
		log.notice "Watch added for " + colorize("[cyan]{#{file}}")
		fs.unwatchFile "#{file}"
		fs.watchFile "#{file}", (curr, prev) ->
			if +curr.mtime isnt +prev.mtime
				log.notice "Saw change in " + colorize("[cyan]{#{file}}")
				runTests (err) ->
					return

# Creates the default spec file.
writeSpecFile = (specFile, prodFile, callback) ->
	libFile = path.basename(prodFile, path.extname(prodFile))
	specContent = ["replace with libfile", "assert = require('assert')", "module.exports =", "\t\"You need to modify this with real test cases and thus it fails until this file is modified.\": ->", "\t\tassert.ok false, \"it will fail\"" ]
	specContent[0] = "#{libFile.replace(/\-/,"_")} = require('#{libFile}')"

	fs.writeFile specFile, specContent.join('\n'), 'utf8', (err) ->
		log.notice "Created specfile '#{specFile}'"
		callback err, null